# -*- coding: utf-8 -*-
"""
Created on Wed May  6 10:56:38 2020

@author: xavier.mouy
"""

import pandas as pd
import os

class DeploymentInfo():
    """
    A class to handle acoustic deployment metadadata .

    Object carrying deployment metadata that can be used for example to populate
    metadata fields in Annotation or Measurement objects.

    Attributes
    ----------
    data : pandas DataFrame
        DataFranme with deploymnent information.

    Methods
    -------
    write_template(filepath)
        Create an empty template csv file with the proper headers.
    read(filepath)
        Populates the DeploymentInfo object with the information from a csv
        file. The csv file must follow the samestructure as the one created by
        the method write_template.

    """

    def __init__(self):
        """
        Initialize object with empty .data attribute.

        Returns
        -------
        None.

        """
        self.data =[]

    def write_template(self, filepath):
        """
        Create a blank deployment file.

        Create an empty template csv file with the proper headers. The created
        csv file has only the headers and an operator must fill in all the
        deployment information manually. Once filled in, this file can be used
        by the DeploymentInfo.read method

        Parameters
        ----------
        filepath : str
            path and name of the deployment csv file to create.

        Returns
        -------
        None. Write a blank csv deployment file that an operator can fill in.

        """
        if os.path.isfile(filepath):
            raise ValueError('File already exists.')
        metadata = pd.DataFrame({
            'audio_channel_number': [],
            'UTC_offset': [],
            'sampling_frequency': [],
            'bit_depth': [],
            'mooring_platform_name': [],
            'recorder_type': [],
            'recorder_SN': [],
            'hydrophone_model': [],
            'hydrophone_SN': [],
            'hydrophone_depth': [],
            'location_name': [],
            'location_lat': [],
            'location_lon': [],
            'location_water_depth': [],
            'deployment_ID': [],
            'deployment_date':[],
            'recovery_date':[],
            })
        metadata.to_csv(filepath,
                        sep=',',
                        encoding='utf-8',
                        header=True,
                        index=False,
                        )

    def read(self, filepath):
        """
        Read metadata information from csv file.

        Load data from a csv file containing the deployment metadat information
        and populated the data attribute of the DeploymentInfo object. The csv
        file must follow the same headers and data format as the csv file
        template generated by DeploymentInfo.write_template.

        Parameters
        ----------
        filepath : str
            Path of the csv file to read.

        Returns
        -------
        None. Populates the pandas dataframe in teh .data attribute of the
        DeploymentInfo object.

        """
        df = pd.read_csv(filepath,
                         delimiter=',',
                         #header=None,
                         skiprows=0,
                         na_values=None,
                         )
        self.data = df
        return df




# -*- coding: utf-8 -*-
"""
Created on Fri Feb  7 15:27:39 2020

@author: xavier.mouy
"""
from .detector_builder import BaseClass
from ecosound.core.spectrogram import Spectrogram
from ecosound.core.annotation import Annotation

from scipy import signal, ndimage
from datetime import datetime
import numpy as np
import pandas as pd
import cv2
import uuid
from numba import njit
import dask
import dask_image.ndfilters
import dask.array

class BlobDetector(BaseClass):
    """Blob detector.

    A detector to find transient events in a spectrogram object. The local
    variance is calculated in the spectrogram for each time-frequency bin using
    a local area (e.g. kernel) defined by 'kernel_duration' and
    'kernel_bandwidth'. Bins or the spectrogram with a local variance less than
    'threshold' are set to zero, while all the bins greater than 'threshold'
    are set to one. The Moors Neighborhood  algorithm is then used to define
    the time and frequency boudaries of the adjacent spectrogram bins taht
    equal one. All detections with a duration less than 'duration_min' and a
    bandwidth less than 'bandwidth_min' are discarded.

    The BlobDetector detector must be instantiated using the DetectorFactory
    with the positional argument 'BlobDetector':

    from ecosound.detection.detector_builder import DetectorFactory
    detector = DetectorFactory('BlobDetector', args)

    Attributes
    ----------
    name : str
        Name of the detector
    version : str
        Version of the detector
    kernel_duration : float
        Duration of the kernel, in seconds.
    kernel_bandwidth : float
        Bandwidth of teh kernel, in Hz.
    threshold : float
        Variance threshold for teh binarization.
    duration_min : float
        Minimum duration of detection accepted,in seconds.
    bandwidth_min : float
        Minimum bandwidth of detection accepted,in seconds.

    Methods
    -------
    run(spectro, debug=False)
        Run the detector on a spectrogram object.
    """

    detector_parameters = ('kernel_duration',
                           'kernel_bandwidth',
                           'threshold',
                           'duration_min',
                           'bandwidth_min')

    def __init__(self, *args, **kwargs):
        """
        Initialize the detector.

        Parameters
        ----------
        *args : str
            Do not use. Only used by the DetectorFactory.
        kernel_duration : float
            Duration of the kernel, in seconds.
        kernel_bandwidth : float
            Bandwidth of teh kernel, in Hz.
        threshold : float
            Variance threshold for the binarization.
        duration_min : float
            Minimum duration of detection accepted,in seconds.
        bandwidth_min : float
            Minimum bandwidth of detection accepted,in seconds.

        Returns
        -------
        None. Detector object.

        """
        # Initialize all detector parameters to None
        self.__dict__.update(dict(zip(self.detector_parameters,
                                      [None]*len(self.detector_parameters))))
        # Unpack kwargs as detector parameters if provided on instantiation
        self.__dict__.update(**kwargs)

    @property
    def name(self):
        """Return name of the detector."""
        detector_name = 'BlobDetector'
        return detector_name

    @property
    def version(self):
        """Return version of the detector."""
        version = '0.1'
        return version

    def _prerun_check(self, spectrogram):
        """Run several verifications before the run."""
        # check that all required arguments are defined
        if True in [self.__dict__.get(keys) is None for keys in self.detector_parameters]:
            raise ValueError('Not all detector parameters have been defined.'
                             + ' Required parameters: '
                             + str(self.detector_parameters))
        # check that spectrogram is a spectrogram class
        if not isinstance(spectrogram, Spectrogram):
            raise ValueError('Input must be an ecosound Spectrogram object'
                             + '(ecosound.core.spectrogram).')

    def _plot_matrix(self, Matrix, title):
        """Plot spectyrogram matrix when in debug mode."""
        import matplotlib.pyplot as plt
        fig, ax = plt.subplots(
            figsize=(16, 4),
            sharex=True)
        ax.pcolormesh(Matrix, cmap='jet')
        ax.set_title(title)

    def run(self, spectro, start_time=None, use_dask=False, dask_chunks=(1000,1000), debug=False):
        """Run detector.

        Runs the detector on the spectrogram object.

        Parameters
        ----------
        spectro : Spectrogram
            Spectrogram object to detect from.
        debug : bool, optional
            Displays binarization results for debugging purpused.The default
            is False.
        start_time : datetime.datetime, optional
            Start time/date of the signal being processed. If defined, the
            fields 'time_min_date' and 'time_max_date' of the detection
            annotation object are populated. The default is None.
        use_dask, bool, optional
            If True, runs the detector in parallel using Dask. The default is
            False.
        dask_chunks, tuple -> (int, int), optional
            Tuple of two int defining the size of the spectrogram chunks to use
            for the parallel processing: dask_chunks=(number of frequency bins,
             number of time bbins). Only used in use_dask is True. The default
            is (1000, 1000).

        Returns
        -------
        detec : Annotation
            Annotation object with the detection results.

        """
        # Pre-run verifications
        self._prerun_check(spectro)
        # Convert units to spectrogram bins
        kernel_duration = max(
            round(self.kernel_duration/spectro.time_resolution), 1)
        kernel_bandwidth = max(
            round(self.kernel_bandwidth/spectro.frequency_resolution), 1)
        duration_min = max(
            round(self.duration_min/spectro.time_resolution), 2)
        bandwidth_min = max(
            round(self.bandwidth_min/spectro.frequency_resolution), 2)
        if debug:
            self._plot_matrix(spectro.spectrogram, 'Spectrogram matrix')
        # # Apply filter
        if use_dask:
            dask_spectro = dask.array.from_array(spectro.spectrogram, chunks=dask_chunks)
            Svar = dask_image.ndfilters.generic_filter(dask_spectro,
                                                       calcVariance2D,
                                                       size=(kernel_bandwidth, kernel_duration),
                                                       mode='mirror')
            Svar = Svar.compute()
        else:
            Svar = ndimage.generic_filter(spectro.spectrogram,
                                          calcVariance2D,
                                          (kernel_bandwidth, kernel_duration),
                                          mode='mirror')

        if debug:
            self._plot_matrix(np.log(Svar), 'Local variance matrix')
        # binarization
        Svar[Svar < self.threshold] = 0
        Svar[Svar > 0] = 1
        if debug:
            self._plot_matrix(Svar, 'Binarized spectrogram matrix')

        #new
        #Svar = cv2.cvtColor(cv2.UMat(Svar), cv2.COLOR_RGB2GRAY)
        # Define contours
        Svar_gray = cv2.normalize(src=Svar,
                                  dst=None,
                                  alpha=0,
                                  beta=255,
                                  norm_type=cv2.NORM_MINMAX,
                                  dtype=cv2.CV_8UC1)
        (cnts, hierarchy) = cv2.findContours(Svar_gray.copy(),
                                                  cv2.RETR_EXTERNAL,
                                                  cv2.CHAIN_APPROX_SIMPLE)
        # loop over the contours
        isdetec = False
        t1 = []
        t2 = []
        fmin = []
        fmax = []
        for c in cnts:
            # Compute the bounding box for the contour
            (x, y, w, h) = cv2.boundingRect(c)
            # if the contour is too small, ignore it
            if w < duration_min or h < bandwidth_min:
                continue
            else:
                isdetec = True
                # box coord
                t1.append(x)
                t2.append(x+w-1)
                fmin.append(y)
                fmax.append(y+h-1)
        # Insert results in an Annotation object
        detec = Annotation()
        detec.data['time_min_offset'] = [t*spectro.time_resolution for t in t1]
        detec.data['time_max_offset'] = [t*spectro.time_resolution for t in t2]
        #detec.data['frequency_min'] = [f*spectro.frequency_resolution for f in fmin]
        #detec.data['frequency_max'] = [f*spectro.frequency_resolution for f in fmax]
        #detec.data['frequency_min'] = [(f*spectro.frequency_resolution)+spectro.axis_frequencies[0] for f in fmin]
        #detec.data['frequency_max'] = [(f*spectro.frequency_resolution)+spectro.axis_frequencies[0] for f in fmax]
        detec.data['frequency_min'] = [spectro.axis_frequencies[f] for f in fmin]
        detec.data['frequency_max'] = [spectro.axis_frequencies[f] for f in fmax]
        detec.data['duration'] = detec.data['time_max_offset'] - detec.data['time_min_offset']
        detec.data['from_detector'] = True
        detec.data['software_name'] = self.name
        detec.data['software_version'] = self.version
        detec.data['entry_date'] = datetime.now()
        detec.data['uuid'] = detec.data.apply(lambda _: str(uuid.uuid4()), axis=1)
        if start_time:
            detec.data['time_min_date']= pd.to_datetime(start_time + pd.to_timedelta(detec.data['time_min_offset'], unit='s'))
            detec.data['time_max_date']= pd.to_datetime(start_time + pd.to_timedelta(detec.data['time_max_offset'], unit='s'))
        return detec

@njit()
def calcVariance2D(buffer):
    """Calculate the 2D variance."""
    return np.var(buffer)
    # return np.median(buffer.ravel())
    # return np.mean(buffer.ravel())



# -*- coding: utf-8 -*-
"""
Created on Mon Jun  1 11:24:56 2020

@author: xavier.mouy
"""

import sys
sys.path.append("..")  # Adds higher directory to python modules path.
from ecosound.core.measurement import Measurement
from ecosound.core.annotation import Annotation
import pandas as pd

"""
Gathers measuremenst for all annotations and noise. Merges into a single dataset,
and re-label classes to create a 2-class dataset 'FS' vs 'NN'.

"""

# Define input and output files
annot_file = r'C:\Users\xavier.mouy\Documents\PhD\Projects\Dectector\results\dataset_annotations_only.nc'
noise_file = r'C:\Users\xavier.mouy\Documents\PhD\Projects\Dectector\results\Noise_dataset'
outfile=r'C:\Users\xavier.mouy\Documents\PhD\Projects\Dectector\results\dataset_FS-NN_modified_20201105145300.nc'

# Load measurements
meas_annot = Measurement()
meas_annot.from_netcdf(annot_file)
meas_noise = Measurement()
meas_noise.from_netcdf(noise_file)

## Label noise measurement as 'NN'
meas_noise.insert_values(label_class='NN')
print(meas_noise.summary())

## relabel annotations that are not 'FS' as 'NN'
print(meas_annot.summary())
meas_annot.data['label_class'].replace(to_replace=['', 'ANT','HS','KW','UN'], value='NN', inplace=True)
print(meas_annot.summary())

## merge the 2 datasets
meas_NN_FS = meas_noise + meas_annot
print(meas_NN_FS.summary())

## Save dataset to nc file
meas_NN_FS.to_netcdf(outfile)




# -*- coding: utf-8 -*-
"""
Created on Fri Nov  6 15:35:40 2020

@author: xavier.mouy
"""

import sys
sys.path.append("..")  # Adds higher directory to python modules path.
from ecosound.core.measurement import Measurement
import os
import pickle
import numpy as np

#indir = r'C:\Users\xavier.mouy\Documents\PhD\Projects\Dectector\DFO_RCA_run\RCA_in_April_July2019_1342218252'
#outdir = r'C:\Users\xavier.mouy\Documents\PhD\Projects\Dectector\DFO_RCA_run2\RCA_in_April_July2019_1342218252'

indir = r'E:\xavier.mouy_20201125\Documents\PhD\Projects\Dectector\DFO_RCA_run\RCA_Out_Oct17_Dec3_2018_67391492'
outdir = r'F:\results\Run_3_large_dataset_balanced\RCA_Out_Oct17_Dec3_2018_67391492'

#indir = r'E:\xavier.mouy_20201125\Documents\PhD\Projects\Dectector\DFO_RCA_run\RCA_In_Jan_April_2019_67674121'
#outdir = r'F:\results\Run_3_large_dataset_balanced\RCA_In_Jan_April_2019_67674121'


classif_model_file = r'C:\Users\xavier.mouy\Documents\PhD\Projects\Detector\results\Classification\20201208T223420_RF_large-noise-dataset\RF50_model_20201208T223420.sav'
file_ext = '.nc'


# laod classif model
classif_model = pickle.load(open(classif_model_file, 'rb'))
features = classif_model['features']
model = classif_model['model']
Norm_mean = classif_model['normalization_mean']
Norm_std = classif_model['normalization_std']
classes_encoder = classif_model['classes']

# loops thrugh each file
files_list = os.listdir(indir) # list of files
for file in files_list:
    if os.path.isfile(os.path.join(indir, file)) & file.endswith(file_ext):
        if os.path.isfile(os.path.join(outdir,file)) is False:
            # load file
            print(file)
            meas = Measurement()
            meas.from_netcdf(os.path.join(indir,file))
            # reclassify
            data = meas.data
            n1=len(data)
            # drop observations/rows with NaNs
            data = data.replace([np.inf, -np.inf], np.nan)
            data.dropna(subset=features, axis=0, how='any', thresh=None, inplace=True)
            n2=len(data)
            print('Deleted observations (due to NaNs): ' + str(n1-n2))
            # Classification - predictions
            X = data[features]
            X = (X-Norm_mean)/Norm_std
            pred_class = model.predict(X)
            pred_prob = model.predict_proba(X)
            pred_prob = pred_prob[range(0,len(pred_class)),pred_class]
            # Relabel
            for index, row in classes_encoder.iterrows():
                pred_class = [row['label'] if i==row['ID'] else i for i in pred_class]
            # update measurements
            data['label_class'] = pred_class
            data['confidence'] = pred_prob
            # sort detections by ascending start date/time
            data.sort_values('time_min_offset',axis=0,ascending=True,inplace=True)
            # save result as NetCDF file
            print('Saving')
            meas.data = data
            meas.to_netcdf(os.path.join(outdir,file))
